module Ncrecon = struct

open Utils
open Ncmatrix
open Ncproof
open Extclause


(* Given a theorem `|- p1 \/ ... \/ pn` as well as theorems
   `p1 |- c`, ..., `pn |- c`, return `|- c`.
*)
let rec DISJ_CASES_LIST th prfs =
  assert (List.length prfs = List.length (disjuncts (concl th)));
  assert (List.for_all2 List.mem (disjuncts (concl th)) (List.map hyp prfs));
  (* the conclusion of all proofs has to be the same *)
  assert (let c = concl (List.hd prfs) in
    List.fold_left (fun acc x -> acc && concl x = c) true prfs);
  match prfs with
    [] -> failwith "DISJ_CASES_LIST"
  | [p] -> MP (DISCH (concl th) p) th
  | [p1; p2] -> DISJ_CASES th p1 p2
  | p :: ps -> DISJ_CASES th p (DISJ_CASES_LIST (ASSUME (snd (dest_disj (concl th)))) ps);;

assert (DISJ_CASES_LIST (ASSUME `a \/ b \/ c`)
  [ TAUT `a ==> c \/ b \/ a` |> UNDISCH
  ; TAUT `b ==> c \/ b \/ a` |> UNDISCH
  ; TAUT `c ==> c \/ b \/ a` |> UNDISCH] |> concl = `c \/ b \/ a`);;


let CONTRADICTION x y =
  let xc, yc = concl x, concl y in
  if !verbose then Format.printf "contra: %s with %s\n%!" (string_of_term xc) (string_of_term yc);
  if is_neg xc && not (is_neg yc) then MP (NOT_ELIM x) y
  else if is_neg yc && not (is_neg xc) then MP (NOT_ELIM y) x
  else failwith "no contradiction found"

assert (concl (CONTRADICTION (ASSUME `~p`) (ASSUME `p:bool`)) = `F`);;
assert (concl (CONTRADICTION (ASSUME `p:bool`) (ASSUME `~p`)) = `F`);;


let flip_neg tm = if is_neg tm then dest_neg tm else mk_neg tm


let nth_disj j th =
  if !verbose then Format.printf "nth_disj %d %s\n%!" j (string_of_term (concl th));
  concl th |> disjuncts |> List.map ASSUME |> List.nth_rest j

let nth_conj i th =
  if !verbose then Format.printf "nth_conj %d %s\n%!" i (string_of_term (concl th));
  List.nth (CONJUNCTS th) i


let inst_var sub ty v =
  let ftm = Substarray.inst_tm sub (Fvar v) in
  try Mapping.hol_of_term ty ftm
  with Failure e -> failwith ("inst_var: " ^ e)

let spec_vars sub vs th =
  if !verbose then
    Format.printf "spec_vars: |vars| = %d, th = %s\n%!"
    (List.length vs) (string_of_term (concl th));
  let tys = concl th |> strip_forall |> fst |> List.map type_of in
  assert (List.length tys = List.length vs);
  List.fold_left2 (fun th ty v -> SPEC (inst_var sub ty v) th)
    th tys vs


type recon_data =
{ problem : thm
; mapping : (int * int) list
; sub : Substarray.t
}

let rec clause_index_mapping cl = cl |>
  List.mapi (fun idx -> function
      Lit _ -> []
    | Mat (j, mat) -> (j, idx) :: matrix_index_mapping mat) |>
  List.concat
and matrix_index_mapping mat = mat |>
  List.mapi (fun idx ((i, v), cl) -> (i, idx) :: clause_index_mapping cl) |>
  List.concat

let mk_recon_data (problem, mat, sub) =
  { problem = problem
  ; mapping = matrix_index_mapping mat
  ; sub = sub
  }

let prefix_problem data =
  let rel i = List.assoc i data.mapping in
  List.fold_left (fun th ((i, v), j) ->
    th |> nth_conj (rel i) |> spec_vars data.sub v |> nth_disj (rel j) |> fst)
    data.problem


let rec eat_proofs data =
  List.fold_map (fun (lem, prfs) th ->
    match prfs with
      prf :: prfs ->
        let bot = bot_of_proof data lem th prf in
        ((concl th, bot) :: lem, prfs), bot
    | [] -> failwith "eat_proofs"
  )

and bot_of_clause_ext data ((i, v), (matL, matR), lmext) acc th =
  let i' = List.assoc i data.mapping in
  assert (i' = List.length matL);
  assert (List.length (CONJUNCTS th) = 1 + List.length (matL @ matR));
  bot_of_litmat_ext data acc (nth_conj i' th |> spec_vars data.sub v) lmext
and bot_of_litmat_ext data acc thi = function
    Litext ((claL, claR), claC) ->
      let j = List.length claL in
      let thj, (thL, thR) = nth_disj j thi in
      let acc1, botsL = eat_proofs data acc  thL in
      let acc2, botsR = eat_proofs data acc1 thR in
      if !verbose then
        Format.printf "bot_of_litmat_ext: thi = %s\n%!" (string_of_thm thi);
      assert (let tmj = concl thj in not (is_conj tmj || is_disj tmj));
      let botj = CONTRADICTION thj (concl thj |> flip_neg |> ASSUME) in
      acc2, DISJ_CASES_LIST thi (botsL @ botj :: botsR)
  | Matext (j, (claL, claR), clext) ->
      let j' = List.assoc j data.mapping in
      assert (j' = List.length claL);
      let thj, (thL, thR) = nth_disj j' thi in
      let acc1, botj  = bot_of_clause_ext data clext acc thj in
      let acc2, botsL = eat_proofs data acc1 thL in
      let acc3, botsR = eat_proofs data acc2 thR in
      acc3, DISJ_CASES_LIST thi (botsL @ botj :: botsR)

and bot_of_proof data lem th = function
    Mat (j, mat), Decomposition (cla, ((i, k), v), prfs) ->
      let i' = List.assoc i data.mapping in
      assert (i' < List.length mat);
      assert (snd (List.nth mat i') = cla);
      if !verbose then
        Format.printf "Decomposition %d: %s\n%!" i' (string_of_term (concl th));
      let thi = nth_conj i' th |> spec_vars data.sub v in
      let djs = disjuncts (concl thi) in
      assert (List.length djs = List.length prfs);
      List.map ASSUME djs |> eat_proofs data (lem, prfs)
      |> snd |> DISJ_CASES_LIST thi
  | Lit lit, prf ->
      let liti = Mapping.hol_of_literal (Substarray.inst_lit data.sub lit) in
      if !verbose then
        Format.printf "bot_of_proof: lit = %s\n%!" (string_of_term liti);
      assert (liti = concl th);
      begin match prf with
        Reduction -> flip_neg liti |> ASSUME |> CONTRADICTION th
      | Lemma -> assert (List.mem_assoc liti lem); List.assoc liti lem
      | Extension (_, (prefix, postfix), prfs) ->
          let prefix_th = prefix_problem data prefix in
          if !verbose then
            Format.printf "prefix_th = %s\n%!" (string_of_thm prefix_th);
          let (_, prfs'), bot =
            bot_of_clause_ext data postfix (lem, prfs) prefix_th in
          assert (prfs' = []);
          assert (List.mem (concl th) (hyp bot));
          MP (DISCH (concl th) bot) th
      | _ -> failwith "bot_of_proof: not a well-formed proof"
      end
  | _ -> failwith "bot_of_proof: not a well-formed proof"

end
