module Ncsearch = struct

open Cop
open Ncmatrix
open Extclause
open Strom
open Fol_term
open Utils


let unify_vars (sub, off) l1 l2 =
  try Some (Subst.unify_tms sub (List.map var l1) (List.map var l2), off)
  with Subst.Unify -> None

let unify_rename (sub, off) args1 (clext, args2, _, vars) =
  try Some (
    if vars = 0 then ((Subst.unify_tms sub args1 args2, off), clext)
    else
      (Subst.unify_tms_off off sub args1 args2, off + vars),
      offset_clause_ext off clext)
  with Subst.Unify -> None


(* epsilon operator on lists *)
let pick f l =
  let rec go acc = function
    x :: xs -> begin match f x with
      Some y -> (y, fun z -> List.rev_append acc (z::xs))
    | None -> go (x::acc) xs end
  | [] -> failwith "pick: no element found" in
  go [] l

let nth_clause i ((((j, k), v), _) as c) = if i = j then Some c else None
let nth_matrix i = function
    Lit _ -> None
  | Mat ((j, k), _ as m) -> if i = j then Some m else None

let unique f l = List.length (List.filter_map f l) = 1

let rec prove_ec k sub ((i, v), _, lmext as clext) mi pi =
  assert (unique (nth_clause i) mi);
  assert (List.length pi = k);
  (*if !verbose then Format.printf "prove_ec: ClaB = %s\n%!" (Ncprint.string_of_matrix (fst (matBC_of_clause_ext clext)));*)
  let ((((i, k1), v1), cla1), miAB) = pick (nth_clause i) mi in
  let alt1 () =
    match lmext with Matext (j, _, clext') when List.mem (j, k1) pi ->
      begin match unify_vars sub v v1 with None -> nil | Some sub2 ->
        let (((j, k1'), mi2), claDE) = pick (nth_matrix j) cla1 in
        assert (unique (nth_matrix j) cla1);
        assert (k1 = k1');
        prove_ec k sub2 clext' mi2 pi |> Strom.map
          (fun (sub3, (prefix, postfix), claB1, mi3) ->
            (sub3, (((i, v1), j) :: prefix, postfix), claB1, miAB (((i, k1), v1), claDE (Mat ((j, k1), mi3))))
          )
      end
    | _ -> nil
  and alt2 () =
    if List.mem (i, k1) pi && v = [] then nil
    else
      let (claB, claC) = claBC_of_litmat_ext lmext in
      let index cla = (((i, k), v), copy_clause k cla) in
      cons ((sub, ([], clext), index claB, miAB (index claC))) nil in
  Strom.(append (cache alt1) (cache alt2))

let cut p xs ys = Strom.(if p then cut xs ys else append xs ys)


open Ncproof
open Print
open Ncprint

let rec prove_lit opt sub (mi, path, pi, lem, lim) lit =
  (*if verbosenc then Format.printf "%s\n%!" (string_of_lit (pred_of_lit lit));*)
  if !verbose then Format.printf "Lit: %s\n%!" (string_of_lit (Subst.inst_lit (fst sub) lit));
  if !verbose then Format.printf "Path: %s\n%!" (string_of_lits path);
  if !verbose then Format.printf "Lemmas: %s\n%!" (string_of_lits lem);
  let neglit = negate_lit lit
  and k = lazy (List.length pi) in
  let lemmas =
    if List.exists (Subst.eq sub lit) lem then (if !verbose then Format.printf "lemma\n%!"; cons (sub, lem, Lemma) nil) else nil
  and reductions = Strom.of_list path |> Strom.filter_map
    (fun p -> if !verbose then Format.printf "Reduction try %s\n%!" (string_of_lit p);
      match Subst.unify sub neglit p with
        Some sub1 -> if !verbose then Format.printf "Reduction works\n%!";
        Some (sub1, lem, Reduction)
      | None -> None)
  and extensions = Ncdatabase.db_entries neglit
    |> Strom.of_list
    |> Strom.concat_map (fun ((_, args, ground, vars) as contra) ->
      if !verbose then Format.printf "Extension try (for lit %s, args (%s), lim %d, vars %d)\n%!" (string_of_lit (Subst.inst_lit (fst sub) lit)) (String.concat "," (List.map string_of_fterm args)) lim vars;
      if lim < 0 && not ground then nil
      else match unify_rename sub (snd lit) contra with
        Some (sub1, clext) ->
          if !verbose then Format.printf "Extension works (for lit %s, lim %d)\n%!" (string_of_lit (Subst.inst_lit (fst sub1) lit)) lim;
          if !verbose then Format.printf "Old/new offsets: %d/%d\n%!" (snd sub) (snd sub1);
          (*incr Stats.infer;
          Format.printf "ClaB: %s\n%!" (Ncprint.string_of_matrix (fst (matBC_of_clause_ext clext)));*)
          prove_ec (Lazy.force k) sub1 clext mi pi
          |> Strom.concat_map (fun (sub2, position, ((i, v), claB1), mi1) ->
            if !verbose then Format.printf "Extension clause %s found\n%!" (string_of_nclause claB1);
            prove_clause opt sub2 (mi1, lit :: path, i::pi, lem, lim - 1) claB1
            |> Strom.map (fun (sub3, prfs) ->
(*
              if verbosenc then Format.printf "Lit__: %s\n%!" (string_of_lit (inst_lit (fst sub3) lit));
              if verbosenc then Format.printf "ClaB1: %s\n%!" (string_of_nclause claB1);
*)
              (sub3, lit :: lem, Extension (Lit neglit :: claB1, position, prfs))))
      | None -> nil) in
  cut (Subst.ground_lit (fst sub) lit)
  (cut opt.cut1 lemmas (cut opt.cut2 reductions (cut opt.cut3 extensions nil))) nil
  |> Strom.map (fun (sub1, lem1, prf1) -> (sub1, lem1, (Lit lit, prf1)))
(* decomposition rule *)
and prove_mat opt sub (mi, path, pi, lem, lim) (j, mat1) =
  if !verbose then Format.printf "prove_mat\n%!";
  Strom.of_list mat1 |> Strom.concat_map
    (fun ((i, _) as idx, cla1) ->
      if !verbose then Format.printf "Decomposition chose: %s\n%!" (string_of_nclause cla1);
      prove_clause opt sub (mi, path, i::j::pi, lem, lim) cla1
      |> Strom.map (fun (sub1, prf1) -> (sub1, lem, (Mat (j, mat1), Decomposition (cla1, idx, prf1))))
    )
and prove_litmat opt sub st = map_litmat (prove_lit opt sub st) (prove_mat opt sub st)
and prove_clause opt sub (mi, path, pi, lem, lim) = function
  lm :: cla ->
    let check = List.exists (fun x -> List.exists (Subst.eq sub x) path) in
    let regularity =
      if opt.improved_regularity then check (clause_lits (lm :: cla))
      else (match lm with Mat _ -> false | Lit l -> check (l :: clause_lits cla)) in
    if regularity then (if !verbose then Format.printf "regularity\n%!"; nil)
    else prove_litmat opt sub (mi, path, pi, lem, lim) lm |> Strom.concat_map
      (fun (sub1, lem1, prf1) ->
        prove_clause opt sub1 (mi, path, pi, lem1, lim) cla
        |> Strom.map (fun (sub2, prf2) -> sub2, prf1 :: prf2)
      )
  (* axiom *)
  | [] -> if !verbose then Format.printf "axiom\n%!"; cons (sub, []) nil

end
