(* Interaction with GNU make jobserver *)
(* Copyright (C) 2017 by Michael Färber under
   Version 3 of the GNU GPL or any later version. *)

#load "str.cma";;
#load "unix.cma";;

unset_jrh_lexer;;

module Jobserver =
struct

(* Throw Not_found if no MAKEFLAGS passed, e.g. if make is run with -j1 *)
let get_fds () =
  let flags = Unix.getenv "MAKEFLAGS" in
  (*Format.printf "flags = %s\n%!" flags;*)

  (* in GNU make 4.2, "jobserver-auth" is used,
     whereas before,  "jobserver-fds" is used
     be compatible with both *)
  let r = Str.regexp "--jobserver-.*=\\([0-9]+\\),\\([0-9]+\\)" in
  ignore (Str.search_forward r flags 0);
  let rfds = Str.matched_group 1 flags in
  let wfds = Str.matched_group 2 flags in
  (*Format.printf "rfd = %s, wfd = %s\n%!" rfds wfds;*)

  (* OCaml's file_descr type is just int *)
  let rfd = (Obj.magic (int_of_string rfds) : Unix.file_descr)
  and wfd = (Obj.magic (int_of_string wfds) : Unix.file_descr) in
  rfd, wfd

(* wait with flags for finished processes, removing them from list of PIDs *)
let rec waitpids flags pids =
  if pids = [] then []
  else match Unix.waitpid flags (-1) with
      (0, _) -> pids
    | (pid, status) ->
        assert (status = Unix.WEXITED 0);
        assert (List.mem pid pids);
        waitpids flags (List.filter ((<>) pid) pids)

let start_job (rfd, wfd) f =
  let buff = " " in
  let read = Unix.read rfd buff 0 1 in
  assert (read = 1);
  match Unix.fork () with
    0 ->
      f ();
      let written = Unix.write wfd buff 0 1 in
      assert (written = 1);
      exit 0
  | pid -> pid

(* start a new job and remove PIDs of every process that finished in meanwhile,
   to prevent zombie processes that fill up the process table *)
let add_job fds f pids = start_job fds f :: waitpids [Unix.WNOHANG] pids

(* wait for remaining processes to finish *)
let finish_jobs pids = let pids' = waitpids [] pids in assert (pids' = [])

let iter fds f =
  let rec iter pids = function
      [] -> finish_jobs pids
    | x :: xs -> iter (add_job fds (fun () -> f x) pids) xs
  in iter []

let fallback_iter f =
  try let fds = get_fds () in iter fds f
  with Not_found -> List.iter f

end

set_jrh_lexer;;
